import pinocchio as pin
import numpy as np
import copy
import argparse
import time
from functools import partial
from ur_simple_control.visualize.visualize import plotFromDict
from ur_simple_control.util.draw_path import drawPath
from ur_simple_control.dmp.dmp import DMP, NoTC,TCVelAccConstrained 
from ur_simple_control.clik.clik import getClikArgs, getClikController, moveL, moveUntilContact
from ur_simple_control.managers import getMinimalArgParser, ControlLoopManager, RobotManager
from ur_simple_control.basics.basics import moveJPI

#######################################################################
#                            arguments                                #
#######################################################################

def getArgs():
    #######################################################################
    #                          generic arguments                          #
    #######################################################################
    parser = getMinimalArgParser()
    parser = getClikArgs(parser)
    parser.add_argument('--kp', type=float, \
            help="proportial control constant for position errors", \
            default=1.0)
    parser.add_argument('--kv', type=float, \
            help="damping in impedance control", \
            default=0.001)
    parser.add_argument('--cartesian-space-impedance', action=argparse.BooleanOptionalAction, \
        help="is the impedance computed and added in cartesian or in joint space", default=False)

    args = parser.parse_args()
    return args


#######################################################################
#                            control loop                             #
#######################################################################

# feedforward velocity, feedback position and force for impedance
def controller():
    pass

# control loop to be passed to ControlLoopManager
def controlLoopPointImpedance(q_init, controller, robot : RobotManager, i, past_data):
    breakFlag = False
    # TODO rename this into something less confusing
    save_past_dict = {}
    log_item = {}
    q = robot.getQ()
    Mtool = robot.getT_w_e()
    wrench = robot.getWrench()
    log_item['wrench_raw'] = wrench.reshape((6,))
    # deepcopy for good coding practise (and correctness here)
    save_past_dict['wrench'] = copy.deepcopy(wrench)
    # rolling average
    #wrench = np.average(np.array(past_data['wrench']), axis=0)
    # first-order low pass filtering instead
    # beta is a smoothing coefficient, smaller values smooth more, has to be in [0,1]
    wrench = args.beta * wrench + (1 - args.beta) * past_data['wrench'][-1]
    Z = np.diag(np.array([1.0, 1.0, 2.0, 1.0, 1.0, 1.0]))
    #Z = np.diag(np.ones(6))

    wrench = Z @ wrench
    # this jacobian starts at the end of the end-effector.
    # but the wrench is probably modified wrt tip of the end effector
    # because we defined the load in the ur inteface thing.
    # so that might cause some offset.
    # test for this:
    # when i touch the f/t sensor directly, it gives on linear force,
    # when i touch it at the tip of the end-effector, then it give both linear force
    # and torque. so that means there's no memeing going on with internally transforming the wrench.
    # plus the f/t sensor has no way of knowing where the contact is anyway, so you can't account 
    # for this even if you wanted to.
    
    # this jacobian might be wrong
    J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)
    dq = robot.getQd()[:6].reshape((6,1))
    # get joint 
    tau = J.T @ wrench
#    if i % 25 == 0:
#        print(*tau.round(1))
    tau = tau[:6].reshape((6,1))
    # compute control law:
    # - feedback the position 
    # kv is not needed if we're running velocity control
    vel_cmd = args.kp * (q_init[:6].reshape((6,1)) - q[:6].reshape((6,1))) + args.alpha * tau
    #vel_cmd = np.zeros(6)
    robot.sendQd(vel_cmd)

    # immediatelly stop if something weird happened (some non-convergence)
    if np.isnan(vel_cmd[0]):
        breakFlag = True

    # log what you said you'd log
    # TODO fix the q6 situation (hide this)
    log_item['qs'] = q[:6].reshape((6,))
    log_item['dqs'] = dq.reshape((6,))
    log_item['wrench_used'] = wrench.reshape((6,))

    return breakFlag, save_past_dict, log_item



def controlLoopCartesianPointImpedance(Mtool_init, clik_controller, robot, i, past_data):
    breakFlag = False
    # TODO rename this into something less confusing
    save_past_dict = {}
    log_item = {}
    q = robot.getQ()
    Mtool = robot.getT_w_e()
    wrench = robot.getWrench()
    log_item['wrench_raw'] = wrench.reshape((6,))
    save_past_dict['wrench'] = copy.deepcopy(wrench)
    wrench = args.beta * wrench + (1 - args.beta) * past_data['wrench'][-1]
    # good generic values
    #Z = np.diag(np.array([1.0, 1.0, 2.0, 1.0, 1.0, 1.0]))
    # but let's stick to the default for now
    #Z = np.diag(np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]))
    Z = np.diag(np.array([1.0, 1.0, 1.0, 10.0, 10.0, 10.0]))

    wrench = Z @ wrench
    # this jacobian starts at the end of the end-effector.
    # but the wrench is probably modified wrt tip of the end effector
    # because we defined the load in the ur inteface thing.
    # so that might cause some offset.
    # test for this:
    # when i touch the f/t sensor directly, it gives on linear force,
    # when i touch it at the tip of the end-effector, then it give both linear force
    # and torque. so that means there's no memeing going on with internally transforming the wrench.
    # plus the f/t sensor has no way of knowing where the contact is anyway, so you can't account 
    # for this even if you wanted to.
    J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)

    SEerror = Mtool.actInv(Mtool_init)
    err_vector = pin.log6(SEerror).vector 
    v_body_cmd = args.kp * err_vector + args.alpha * wrench
    # now this v_spatial_cmd can be passed to any inverse kinematics algorithm

    vel_cmd = clik_controller(J, v_body_cmd)
    robot.sendQd(vel_cmd)

    # immediatelly stop if something weird happened (some non-convergence)
    if np.isnan(vel_cmd[0]):
        breakFlag = True

    dq = robot.getQd()[:6].reshape((6,1))
    # log what you said you'd log
    # TODO fix the q6 situation (hide this)
    log_item['qs'] = q[:6].reshape((6,))
    log_item['dqs'] = dq.reshape((6,))
    log_item['wrench_used'] = wrench.reshape((6,))

    return breakFlag, save_past_dict, log_item


if __name__ == "__main__":
    #######################################################################
    #                           software setup                            #
    #######################################################################

    args = getArgs()
    if args.debug_prints:
        print("you will get a lot of stuff printed out, as requested")
    clikController = getClikController(args)
    robot = RobotManager(args)
    
    # TODO and NOTE the weight, TCP and inertial matrix needs to be set on the robot
    # you already found an API in rtde_control for this, just put it in initialization 
    # under using/not-using gripper parameters
    # ALSO NOTE: to use this you need to change the version inclusions in
    # ur_rtde due to a bug there in the current ur_rtde + robot firmware version 
    # (the bug is it works with the firmware verion, but ur_rtde thinks it doesn't)
    # here you give what you're saving in the rolling past window 
    # it's initial value.
    # controlLoopManager will populate the queue with these initial values
    save_past_dict = {
            'wrench' : np.zeros(6),
        }
    # here you give it it's initial value
    log_item = {
            'qs' : np.zeros(robot.n_arm_joints),
            'dqs' : np.zeros(robot.n_arm_joints),
            'wrench_raw' : np.zeros(6),
            'wrench_used' : np.zeros(6),
        }
    q_init = robot.getQ()
    Mtool_init = robot.getT_w_e()

    if not args.cartesian_space_impedance:
        controlLoop = partial(controlLoopPointImpedance, q_init, controller, robot)
    else:
        controlLoop = partial(controlLoopCartesianPointImpedance, Mtool_init, clikController, robot)

    loop_manager = ControlLoopManager(robot, controlLoop, args, save_past_dict, log_item)

    #moveJ(args, robot, dmp.pos.reshape((6,)))
    # and now we can actually run
    loop_manager.run()

    #plotFromDict(log_dict, args)
    # plotting is now initiated in stophandler because then we get the plot 
    # even if we end sooner
    loop_manager.stopHandler(None, None)
    # TODO: add some math to analyze path precision

    


