import pinocchio as pin
import numpy as np
import copy
import argparse
from functools import partial
from ur_simple_control.managers import ControlLoopManager, RobotManager
import time
from qpsolvers import solve_qp
import argparse

def getClikArgs(parser):
    """
    getClikArgs
    ------------
    Every clik-related magic number, flag and setting is put in here.
    This way the rest of the code is clean.
    Force filtering is considered a core part of these control loops
    because it's not necessarily the same as elsewhere.
    If you want to know what these various arguments do, just grep 
    though the code to find them (but replace '-' with '_' in multi-word arguments).
    All the sane defaults are here, and you can change any number of them as an argument
    when running your program. If you want to change a drastic amount of them, and
    not have to run a super long commands, just copy-paste this function and put your
    own parameters as default ones.
    """
    #parser = argparse.ArgumentParser(description='Run closed loop inverse kinematics \
    #        of various kinds. Make sure you know what the goal is before you run!',
    #        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    ####################
    #  clik arguments  #
    ####################
    parser.add_argument('--goal-error', type=float, \
            help="the final position error you are happy with", default=1e-2)
    parser.add_argument('--tikhonov-damp', type=float, \
            help="damping scalar in tikhonov regularization", default=1e-3)
    # TODO add the rest
    parser.add_argument('--clik-controller', type=str, \
            help="select which click algorithm you want", \
            default='dampedPseudoinverse', choices=['dampedPseudoinverse', 'jacobianTranspose', 'invKinmQP'])

    ###########################################
    #  force sensing and feedback parameters  #
    ###########################################
    parser.add_argument('--alpha', type=float, \
            help="force feedback proportional coefficient", \
            default=0.01)
    parser.add_argument('--beta', type=float, \
            help="low-pass filter beta parameter", \
            default=0.01)

    ###############################
    #  path following parameters  #
    ###############################
    parser.add_argument('--max-init-clik-iterations', type=int, \
            help="number of max clik iterations to get to the first point", default=10000)
    parser.add_argument('--max-running-clik-iterations', type=int, \
            help="number of max clik iterations between path points", default=1000)
    
    return parser

#######################################################################
#                             controllers                             #
#######################################################################
"""
controllers general
-----------------------
really trully just the equation you are running.
ideally, everything else is a placeholder,
meaning you can swap these at will.
if a controller has some additional arguments,
those are put in via functools.partial in getClikController,
so that you don't have to worry about how your controller is handled in
the actual control loop after you've put it in getClikController,
which constructs the controller from an argument to the file.
"""

def dampedPseudoinverse(tikhonov_damp, J, err_vector):
    qd = J.T @ np.linalg.inv(J @ J.T + np.eye(J.shape[0], J.shape[0]) * tikhonov_damp) @ err_vector
    return qd

def jacobianTranspose(J, err_vector):
    qd = J.T @ err_vector
    return qd

def invKinmQP(J, err_vector):
    # maybe a lower precision dtype is equally good, but faster?
    P = np.eye(J.shape[1], dtype="double")
    # TODO: why is q all 0?
    q = np.array([0] * J.shape[1], dtype="double")
    G = None
    # TODO: extend for orientation as well
    b = err_vector#[:3]
    A = J#[:3]
    # TODO: you probably want limits here
    lb = None
    ub = None
    h = None
    #qd = solve_qp(P, q, G, h, A, b, lb, ub, solver="ecos")
    qd = solve_qp(P, q, G, h, A, b, lb, ub, solver="quadprog")
    return qd

def getClikController(args):
    """
    getClikController
    -----------------
    A string argument is used to select one of these.
    It's a bit ugly, bit totally functional and OK solution.
    we want all of theme to accept the same arguments, i.e. the jacobian and the error vector.
    if they have extra stuff, just map it in the beginning with partial
    NOTE: this could be changed to something else if it proves inappropriate later
    TODO: write out other algorithms
    """
    if args.clik_controller == "dampedPseudoinverse":
        return partial(dampedPseudoinverse, args.tikhonov_damp)
    if args.clik_controller == "jacobianTranspose":
        return jacobianTranspose
    # TODO implement and add in the rest
    #if controller_name == "invKinmQPSingAvoidE_kI":
    #    return invKinmQPSingAvoidE_kI
    #if controller_name == "invKinm_PseudoInv":
    #    return invKinm_PseudoInv
    #if controller_name == "invKinm_PseudoInv_half":
    #    return invKinm_PseudoInv_half
    if args.clik_controller == "invKinmQP":
        return invKinmQP
    #if controller_name == "invKinmQPSingAvoidE_kI":
    #    return invKinmQPSingAvoidE_kI
    #if controller_name == "invKinmQPSingAvoidE_kM":
    #    return invKinmQPSingAvoidE_kM
    #if controller_name == "invKinmQPSingAvoidManipMax":
    #    return invKinmQPSingAvoidManipMax

    # default
    return partial(dampedPseudoinverse, args.tikhonov_damp)


# modularity yo
# past data to comply with the API
# TODO make this a kwarg or whatever to be more lax with this
def controlLoopClik(robot, clik_controller, i, past_data):
    """
    controlLoopClik
    ---------------
    generic control loop for clik (handling error to final point etc).
    in some version of the universe this could be extended to a generic
    point-to-point motion control loop.
    """
    breakFlag = False
    log_item = {}
    # know where you are, i.e. do forward kinematics
    q = robot.getQ()
    # first check whether we're at the goal
    SEerror = robot.data.oMi[robot.JOINT_ID].actInv(robot.Mgoal)
    err_vector = pin.log6(SEerror).vector 
    if np.linalg.norm(err_vector) < robot.args.goal_error:
#      print("Convergence achieved, reached destionation!")
        breakFlag = True
    # pin.computeJointJacobian is much different than the C++ version lel
    # TODO: put in step, havew getter for it
    J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)
    # compute the joint velocities.
    # just plug and play different ones
    qd = clik_controller(J, err_vector)
    robot.sendQd(qd)
    
    # TODO: only print under a debug flag, it's just clutter otherwise
    # we do the printing here because controlLoopManager should be general.
    # and these prints are click specific (although i'm not 100% happy with the arrangement)
#    if not i % 1000:
#        print("pos", robot.data.oMi[robot.JOINT_ID])
#        print("linear error = ", pin.log6(SEerror).linear)
#        print("angular error = ", pin.log6(SEerror).angular)
#        print(" error = ", err_vector.transpose())

    log_item['qs'] = q.reshape((robot.model.nq,))
    # get actual velocity, not the commanded one.
    # although that would be fun to compare i guess
    # TODO: have both, but call the compute control signal like it should be
    log_item['dqs'] = robot.getQd().reshape((robot.model.nq,))
    # we're not saving here, but need to respect the API, 
    # hence the empty dict
    return breakFlag, {}, log_item


def moveUntilContactControlLoop(contact_detecting_force, speed, robot : RobotManager, clik_controller, i, past_data):
    """
    moveUntilContactControlLoop
    ---------------
    generic control loop for clik (handling error to final point etc).
    in some version of the universe this could be extended to a generic
    point-to-point motion control loop.
    """
    breakFlag = False
    # know where you are, i.e. do forward kinematics
    q = robot.getQ()
    # break if wrench is nonzero basically
    #wrench = robot.getWrench()
    # you're already giving the speed in the EE i.e. body frame
    # so it only makes sense to have the wrench in the same frame
    #wrench = robot._getWrenchInEE()
    wrench = robot.getWrench()
    # and furthermore it's a reasonable assumption that you'll hit the thing
    # in the direction you're going in.
    # thus we only care about wrenches in those direction coordinates
    mask = speed != 0.0
    # NOTE: contact getting force is a magic number
    # it is a 100% empirical, with the goal being that it's just above noise.
    # so far it's worked fine, and it's pretty soft too.
    if np.linalg.norm(wrench[mask]) > contact_detecting_force:
        print("hit with", np.linalg.norm(wrench[mask]))
        breakFlag = True
    # pin.computeJointJacobian is much different than the C++ version lel
    J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)
    # compute the joint velocities.
    qd = clik_controller(J, speed)
    robot.sendQd(qd)
    return breakFlag, {}, {}

def moveUntilContact(args, robot, speed):
    """
    moveUntilContact
    -----
    does clik until it feels something with the f/t sensor
    """
    assert type(speed) == np.ndarray 
    clik_controller = getClikController(args)
    # TODO: just make the manager pass the robot or something, this is weird man
    controlLoop = partial(moveUntilContactControlLoop, args.contact_detecting_force, speed, robot, clik_controller)
    # we're not using any past data or logging, hence the empty arguments
    loop_manager = ControlLoopManager(robot, controlLoop, args, {}, {})
    loop_manager.run()
    # TODO: remove, this isn't doing anything
    time.sleep(0.01)
    print("Collision detected!!")

def moveL(args, robot, goal_point):
    """
    moveL
    -----
    does moveL.
    send a SE3 object as goal point.
    if you don't care about rotation, make it np.zeros((3,3))
    """
    #assert type(goal_point) == pin.pinocchio_pywrap.SE3
    robot.Mgoal = copy.deepcopy(goal_point)
    clik_controller = getClikController(args)
    controlLoop = partial(controlLoopClik, robot, clik_controller)
    # we're not using any past data or logging, hence the empty arguments
    log_item = {
            'qs' : np.zeros(robot.model.nq),
            'dqs' : np.zeros(robot.model.nq),
        }
    loop_manager = ControlLoopManager(robot, controlLoop, args, {}, log_item)
    loop_manager.run()
    # TODO: remove, this isn't doing anything
    time.sleep(0.01)
    print("MoveL done: convergence achieved, reached destionation!")


def controlLoopCompliantClik(args, robot : RobotManager, i, past_data):
    """
    controlLoopClik
    ---------------
    generic control loop for clik (handling error to final point etc).
    in some version of the universe this could be extended to a generic
    point-to-point motion control loop.
    """
    breakFlag = False
    log_item = {}
    save_past_dict = {}
    # know where you are, i.e. do forward kinematics
    q = robot.getQ()
    T_w_e = robot.getT_w_e()
    wrench = robot.getWrench()
    # we need to overcome noise if we want to converge
    if np.linalg.norm(wrench) < args.minimum_detectable_force_norm:
        wrench = np.zeros(6)
    save_past_dict['wrench'] = copy.deepcopy(wrench)
    wrench = args.beta * wrench + (1 - args.beta) * past_data['wrench'][-1]
    mapping = np.zeros((6,6))
    mapping[0:3, 0:3] = T_w_e.rotation
    mapping[3:6, 3:6] = T_w_e.rotation
    wrench = mapping.T @ wrench
    #Z = np.diag(np.array([1.0, 1.0, 2.0, 1.0, 1.0, 1.0]))
    Z = np.diag(np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]))
    wrench = Z @ wrench
    #pin.forwardKinematics(robot.model, robot.data, q)
    # first check whether we're at the goal
    SEerror = T_w_e.actInv(robot.Mgoal)
    err_vector = pin.log6(SEerror).vector 
    if np.linalg.norm(err_vector) < robot.args.goal_error:
#      print("Convergence achieved, reached destionation!")
        breakFlag = True
    # pin.computeJointJacobian is much different than the C++ version lel
    J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)
    # compute the joint velocities.
    # just plug and play different ones
    qd = J.T @ np.linalg.inv(J @ J.T + np.eye(J.shape[0], J.shape[0]) * args.tikhonov_damp) @ err_vector
    tau = J.T @ wrench
    #tau = tau[:6].reshape((6,1))
    qd += args.alpha * tau
    robot.sendQd(qd)
    
    log_item['qs'] = q.reshape((robot.model.nq,))
    # get actual velocity, not the commanded one.
    # although that would be fun to compare i guess
    # TODO: have both, but call the compute control signal like it should be
    log_item['dqs'] = robot.getQd().reshape((robot.model.nq,))
    log_item['wrench'] = wrench.reshape((6,))
    log_item['tau'] = tau.reshape((robot.model.nq,))
    # we're not saving here, but need to respect the API, 
    # hence the empty dict
    return breakFlag, save_past_dict, log_item

# add a threshold for the wrench
def compliantMoveL(args, robot, goal_point):
    """
    compliantMoveL
    -----
    does compliantMoveL - a moveL, but with compliance achieved
    through f/t feedback
    send a SE3 object as goal point.
    if you don't care about rotation, make it np.zeros((3,3))
    """
#    assert type(goal_point) == pin.pinocchio_pywrap.SE3
    robot.Mgoal = copy.deepcopy(goal_point)
    clik_controller = getClikController(args)
    controlLoop = partial(controlLoopCompliantClik, args, robot)
    # we're not using any past data or logging, hence the empty arguments
    log_item = {
            'qs' : np.zeros(robot.model.nq),
            'wrench' : np.zeros(6),
            'tau' : np.zeros(robot.model.nq),
            'dqs' : np.zeros(robot.model.nq),
        }
    save_past_dict = {
            'wrench': np.zeros(6),
            }
    loop_manager = ControlLoopManager(robot, controlLoop, args, save_past_dict, log_item)
    loop_manager.run()
    # TODO: remove, this isn't doing anything
    time.sleep(0.01)
    print("compliantMoveL done: convergence achieved, reached destionation!")

def clikCartesianPathIntoJointPath(path, args, robot, \
        clikController, q_init, R, p):
    """
    clikCartesianPathIntoJointPath
    ------------------------------
    functionality
    ------------
    Follows a provided Cartesian path,
    creates a joint space trajectory for it.

    return
    ------
    - joint_space_trajectory to follow the given path.

    arguments
    ----------
    - path:
      --> cartesian path given in task frame
    TODO: write asserts for these arguments
    - args:
      --> all the magic numbers used here better be here
    - clikController:
      --> clik controller you're using
    - robot:
      --> RobotManager instance (needed for forward kinematics etc)
    TODO: maybe it's better design to expect path in body frame idk
          the good thing here is you're forced to think about frames.
    TODO: make this a kwarg with a neural transform as default.
    - transf_body_task_frame: 
      --> A transformation from the body frame to task frame.
      --> It is assumed that the path was provided in the task frame
          because 9/10 times that's more convenient,
          and you can just pass a neural transform if you're not using it.
    - q_init:
      --> starting point. 
      --> you can movej to it before executing the trajectory,
         so this makes perfect sense to ensure correctness
    """

    transf_body_to_task_frame = pin.SE3(R, p)
    q = copy.deepcopy(q_init)
    pin.forwardKinematics(robot.model, robot.data, q)

    for i in range(len(path)):
        path[i] = transf_body_to_task_frame.act(path[i])
    # TODO remove print, write a test for this instead
    if args.debug_prints:
        print(path)

    # TODO: finish this
    # - pass clik alg as argument
    # - remove magic numbers
    # - give output in the right format
    # skip inital pos tho
    #q = np.array([-2.256,-1.408,0.955,-1.721,-1.405,-0.31, 0.0, 0.0])
    #q = np.array([-2.014, -1.469, 1.248, -1.97, -1.366, -0.327, 0.0, 0.0])
    # this is just init_q right now
    # TODO: make this a flag or something for readability's sake
    n_iter = args.max_init_clik_iterations
    # we don't know how many there will be, so a linked list is 
    # clearly the best data structure here (instert is o(1) still,
    # and we aren't pressed on time when turning it into an array later)
    qs = []
    for goal in path:
        #print("goal", goal)
        #print("T_w_e", robot.data.oMi[robot.JOINT_ID])
        Mgoal = pin.SE3(R, goal)
        for i in range(n_iter):
            # TODO maybe hide pin call with RobotManager call
            pin.forwardKinematics(robot.model, robot.data, q)
            SEerror = robot.data.oMi[robot.JOINT_ID].actInv(Mgoal)
            err_vector = pin.log6(SEerror).vector 
            if np.linalg.norm(err_vector) < args.goal_error:
                if not n_iter == args.max_init_clik_iterations:
                    if args.debug_prints:
                        print("converged in", i)
                    break
            J = pin.computeJointJacobian(robot.model, robot.data, q, robot.JOINT_ID)
            qd = clikController(J, err_vector)
            # we're integrating this fully offline of course
            q = pin.integrate(robot.model, q, qd * robot.dt)
            if (not n_iter == args.max_init_clik_iterations) and (i % 10 == 0):
                qs.append(q[:6])

        # just skipping the first run with one stone
        if n_iter == args.max_init_clik_iterations:
            n_iter = args.max_running_clik_iterations
        else:
            if i == args.max_running_clik_iterations - 1:
                print("DID NOT CONVERGE -- exiting")
                # nothing is moving 
                # and i'm not even using a manager here
                # so no need, right?
                #ControlLoopManager.stopHandler(None, None, None)
                exit()

    ##############################################
    #  save the obtained joint-space trajectory  #
    ##############################################
    qs = np.array(qs)
    # we're putting a dmp over this so we already have the timing ready
    # TODO: make this general, you don't want to depend on other random
    # arguments (make this one traj_time, then put tau0 = traj_time there
    t = np.linspace(0, args.tau0, len(qs)).reshape((len(qs),1))
    joint_trajectory = np.hstack((t, qs))
    # TODO handle saving more consistently/intentionally
    # (although this definitely works right now and isn't bad, just mid)
    # os.makedir -p a data dir and save there, this is ugly
    np.savetxt("./joint_trajectory.csv", joint_trajectory, delimiter=',', fmt='%.5f')
    return joint_trajectory


if __name__ == "__main__": 
    args = get_args()
    robot = RobotManager(args)
    Mgoal = robot.defineGoalPointCLI()
    clik_controller = getClikController(args)
    controlLoop = partial(controlLoopClik, robot, clik_controller)
    # we're not using any past data or logging, hence the empty arguments
    loop_manager = ControlLoopManager(robot, controlLoop, args, {}, {})
    loop_manager.run()
