import numpy as np
import pinocchio as pin
import crocoddyl
from ur_simple_control.managers import getMinimalArgParser, ControlLoopManager, RobotManager
import argparse
from ur_simple_control.basics.basics import jointTrajFollowingPD
from ur_simple_control.visualize.visualize import plotFromDict
import example_robot_data
import time


def get_OCP_args(parser : argparse.ArgumentParser):
    parser.add_argument("--ocp-dt", type=float, default=1e-2, \
                        help="time length between state-control-pair decision variables")
    parser.add_argument("--n-knots", type=int, default=100, \
                        help="number of state-control-pair decision variables")
    parser.add_argument("--stop-at-final", type=int, default=1, \
                        help="the trajectory generated by the OCP will be followed. it might not have finally velocity 0. \
                             if this argument is set to true, the final velocity will be set to 0 (there will be overshoot).")
    return parser
                        

# TODO: use fddp and incorporate torque (i.e. velocity constraints)
#   --> those will correspond to max_qd and acceleration attributes in robotmanager 
def CrocoIKOCP(args, robot : RobotManager, goal : pin.SE3):
    # create torque bounds which correspond to percentage
    # of maximum allowed acceleration 
    # NOTE: idk if this makes any sense, but let's go for it
    #print(robot.model.effortLimit)
    #exit()
    #robot.model.effortLimit = robot.model.effortLimit * (args.acceleration / robot.MAX_ACCELERATION)
    #robot.model.effortLimit = robot.model.effortLimit * 0.01
    #robot.data = robot.model.createData()
    # starting state
    x0 = np.concatenate([robot.getQ(), robot.getQd()])
    state = crocoddyl.StateMultibody(robot.model)
    # command input IS torque 
    actuation = crocoddyl.ActuationModelFull(state)

    # we will be summing 4 different costs
    # first 3 are for tracking, state and control regulation
    runningCostModel = crocoddyl.CostModelSum(state)
    terminalCostModel = crocoddyl.CostModelSum(state)
    # cost 1) u residual (actuator cost)
    uResidual = crocoddyl.ResidualModelControl(state, state.nv)
    uRegCost = crocoddyl.CostModelResidual(state, uResidual)
    # cost 2) x residual (overall amount of movement)
    xResidual = crocoddyl.ResidualModelState(state, x0, state.nv)
    xRegCost = crocoddyl.CostModelResidual(state, xResidual)
    # cost 3) distance to goal -> SE(3) error
    framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
            state,
            # TODO: check if this is the frame we actually want (ee tip)
            # the id is an integer and that's what api wants
            robot.model.getFrameId("tool0"),
            goal.copy(),
            state.nv)
    goalTrackingCost = crocoddyl.CostModelResidual(state, framePlacementResidual)
    # cost 4) final ee velocity is 0 (can't directly formulate joint velocity,
    #         because that's a part of the state, and we don't know final joint positions)
    frameVelocityResidual = crocoddyl.ResidualModelFrameVelocity(
            state,
            robot.model.getFrameId("tool0"),
            pin.Motion(np.zeros(6)),
            pin.ReferenceFrame.WORLD)
    frameVelocityCost = crocoddyl.CostModelResidual(state, frameVelocityResidual)
    # put these costs into the running costs
    runningCostModel.addCost("gripperPose", goalTrackingCost, 1e2)
    runningCostModel.addCost("xReg", xRegCost, 1e-3)
    runningCostModel.addCost("uReg", uRegCost, 1e-3)
    # and add the terminal cost, which is the distance to the goal
    # NOTE: shouldn't there be a final velocity = 0 here?
    terminalCostModel.addCost("gripperPose", goalTrackingCost, 1e2)
    terminalCostModel.addCost("uReg", uRegCost, 1e3)
    terminalCostModel.addCost("velFinal", frameVelocityCost, 1e3)

    # the 4th cost is for defining bounds via costs
    # NOTE: could have gotten the same info from state.lb and state.ub.
    # the first state is unlimited there idk what that means really,
    # but the arm's base isn't doing a full rotation anyway, let alone 2 or more
    xlb = np.concatenate([
        robot.model.lowerPositionLimit,
        -1 * robot.model.velocityLimit])
    xub = np.concatenate([
        robot.model.upperPositionLimit,
        robot.model.velocityLimit])
    bounds = crocoddyl.ActivationBounds(xlb, xub, 1.0)
    xLimitResidual = crocoddyl.ResidualModelState(state, x0, actuation.nu)
    xLimitActivation = crocoddyl.ActivationModelQuadraticBarrier(bounds)
    limitCost = crocoddyl.CostModelResidual(state, xLimitActivation, xLimitResidual)
    runningCostModel.addCost("limitCost", limitCost, 1e3)
    terminalCostModel.addCost("limitCost", limitCost, 1e3)
    # NOTE: i have no idea how to incorporate this into anything
    # we need to add state constraints for things to make sense
    # NOTE!!!: there are no hard inequality constraints on states
    # in what's implemented in crocoddyl. 
    # INSTEAD, you define quadratic barier-type costs for that.
    # there are box constraints on control input though,
    # but i'm pretty sure i'm not using those correctly.
    #x_constraint_lower = x0.copy()
    #x_constraint_upper = x0.copy()
    #x_constraint_lower[:robot.model.nq] = -2 * np.pi
    #x_constraint_upper[:robot.model.nq] = 2 * np.pi
    #x_constraint_lower[robot.model.nq:] = -1 * robot.max_qd
    #x_constraint_upper[robot.model.nq:] = robot.max_qd
    #state_constraint = crocoddyl.ConstraintModelResidual(state, xResidual, x_constraint_lower, x_constraint_upper)

    # Next, we need to create an action model for running and terminal knots. The
    # forward dynamics (computed using ABA) are implemented
    # inside DifferentialActionModelFreeFwdDynamics.
    runningModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeInvDynamics(
            state, actuation, runningCostModel
        ),
        args.ocp_dt,
    )
    terminalModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeInvDynamics(
            state, actuation, terminalCostModel
        ),
        0.0,
    )

    # now we define the problem
    problem = crocoddyl.ShootingProblem(x0, [runningModel] * args.n_knots, terminalModel)
    # and the solver
    solver = crocoddyl.SolverBoxFDDP(problem)
    #solver = crocoddyl.SolverIpopt(problem)
    # TODO: remove / place elsewhere once it works
    solver.setCallbacks([crocoddyl.CallbackVerbose(), crocoddyl.CallbackLogger()])
    # run solve
    # NOTE: there are some possible parameters here that I'm not using
    xs = [x0] * (solver.problem.T + 1)
    us = solver.problem.quasiStatic([x0] * solver.problem.T)

    print("set goal:", goal)
    start = time.time()
    solver.solve(xs, us, 500, False, 1e-9)
    end = time.time()
    print("solved in:", end - start, "seconds")

    #solver.solve()
    # get reference (state trajectory)
    # we aren't using controls because we only have velocity inputs
    xs = np.array(solver.xs)
    qs = xs[:, :robot.model.nq]
    vs = xs[:, robot.model.nq:]
    reference = {'qs' : qs, 'vs' : vs, 'dt' : args.ocp_dt}
    return reference, solver

if __name__ == "__main__":
    parser = getMinimalArgParser()
    parser = get_OCP_args(parser)
    args = parser.parse_args()
    ex_robot = example_robot_data.load("ur5_gripper")
    robot = RobotManager(args)
    # TODO: remove once things work
    robot.max_qd = 3.14 
    print("velocity limits", robot.model.velocityLimit)
    robot.q = pin.randomConfiguration(robot.model)
    goal = pin.SE3.Random()
    goal.translation = np.random.random(3) * 0.4
    reference, solver = CrocoIKOCP(args, robot, goal)
    # we only work within -pi - pi (or 0-2pi idk anymore)
    #reference['qs'] = reference['qs'] % (2*np.pi) - np.pi
    log = solver.getCallbacks()[1]
    crocoddyl.plotOCSolution(log.xs, log.us, figIndex=1, show=True)
    crocoddyl.plotConvergence(log.costs, log.pregs, log.dregs, log.grads, log.stops, log.steps, figIndex=2)
    jointTrajFollowingPD(args, robot, reference)
    reference.pop('dt')
    plotFromDict(reference, args.n_knots + 1, args)
    print("achieved result", robot.getT_w_e())
    display = crocoddyl.MeshcatDisplay(ex_robot)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)
